import contextlib
import tempfile
from pathlib import Path
from typing import Annotated, Any

import torch
from pydantic import BaseModel, BeforeValidator, PlainSerializer, WithJsonSchema
from pydantic_yaml import parse_yaml_file_as, to_yaml_file


def validate_torch_dtype(x: Any) -> torch.dtype:
    match x:
        case torch.dtype():
            return x
        case str() if hasattr(torch, x) and isinstance(getattr(torch, x), torch.dtype):
            return getattr(torch, x)
        case _:
            raise ValueError(f"{x} is not a valid torch.dtype.")


SerializableTorchDtype = Annotated[
    torch.dtype,
    BeforeValidator(lambda x: validate_torch_dtype(x)),
    PlainSerializer(lambda x: str(x).split(".")[1]),
    WithJsonSchema({"type": "string"}, mode="validation"),
    WithJsonSchema({"type": "string"}, mode="serialization"),
]
"""Custom type validator for a `torch.dtype` object.

Accepts `torch.dtype` instances or strings representing a valid dtype from the `torch` package.
Ref: https://docs.pydantic.dev/latest/concepts/types/#custom-types
"""


class LMBuddyConfig(
    BaseModel,
    extra="forbid",
    arbitrary_types_allowed=True,
    validate_assignment=True,
):
    """Base class for all Pydantic configs in the library.

    Defines some common settings used by all subclasses.
    """

    @classmethod
    def from_yaml_file(cls, path: Path | str):
        return parse_yaml_file_as(cls, path)

    def to_yaml_file(self, path: Path | str):
        to_yaml_file(path, self, exclude_none=True)

    @contextlib.contextmanager
    def to_tempfile(self, *, name: str = "config.yaml", dir: str | Path | None = None):
        """Enter a context manager with the config written to a temporary YAML file.

        Keyword Args:
            name (str): Name of the config file in the tmp directory. Defaults to "config.yaml".
            dir (str | Path | None): Root path of the temporary directory. Defaults to None.

        Returns:
            Path to the temporary config file.
        """
        with tempfile.TemporaryDirectory(dir=dir) as tmpdir:
            config_path = Path(tmpdir) / name
            self.to_yaml_file(config_path)
            yield config_path



from pydantic import Field

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.huggingface import DatasetConfig
from lm_buddy.configs.jobs.common import JobConfig
from lm_buddy.configs.vllm import VLLMCompletionsConfig
from lm_buddy.paths import AssetPath


class PrometheusEvaluationConfig(LMBuddyConfig):
    """Parameters specific to Prometheus evaluation."""

    num_answers: int = 3
    max_retries: int = 5
    scores: list = [1, 2, 3, 4, 5]
    min_score: int = 0
    max_score: int = 5
    enable_tqdm: bool = False
    conversation_template: str = "llama-2"
    conversation_system_message: str = "You are a fair evaluator language model."
    storage_path: str | None = None


class PrometheusJobConfig(JobConfig):
    """Configuration for a Prometheus judge evaluation task."""

    prometheus: VLLMCompletionsConfig = Field(
        description="Externally hosted Prometheus judge model."
    )
    dataset: DatasetConfig = Field(
        description="Dataset of text completions to evaluate using the Prometheus judge model."
    )
    evaluation: PrometheusEvaluationConfig = Field(
        default_factory=PrometheusEvaluationConfig,
        description="Settings for the Prometheus evaluation.",
    )

    def asset_paths(self) -> set[AssetPath]:
        paths = {self.dataset.path, self.prometheus.inference.engine}
        return {x for x in paths if x is not None}



from enum import Enum
from pathlib import Path
from urllib.parse import ParseResult, urlparse

import pandas as pd
import wandb

from lm_buddy.paths import PathPrefix


class ArtifactType(str, Enum):
    """Enumeration of artifact types used by the LM Buddy."""

    DATASET = "dataset"
    MODEL = "model"
    TOKENIZER = "tokenizer"
    EVALUATION = "evaluation"


def default_artifact_name(job_name: str, artifact_type: ArtifactType) -> str:
    """A default name for an artifact based on the job name and type."""
    return f"{job_name}-{artifact_type}"


def get_artifact_from_api(artifact_name: str) -> wandb.Artifact:
    """Retrieve an artifact by fully qualified name from the W&B API.

    This does not handle linking the artifact to an active run.
    For that, use `run.use_artifact(artifact_name)`.
    """
    api = wandb.Api()
    return api.artifact(artifact_name)


def get_artifact_directory(
    artifact: wandb.Artifact, *, download_root_path: str | None = None
) -> Path:
    """Get the directory containing the artifact's data.

    If the artifact references data already on the filesystem, simply return that path.
    If not, downloads the artifact (with the specified `download_root_path`)
    and returns the newly created artifact directory path.
    """
    for entry in artifact.manifest.entries.values():
        match urlparse(entry.ref):
            case ParseResult(scheme="file", path=file_path):
                return Path(file_path).parent
    # No filesystem references found in the manifest -> download the artifact
    download_path = artifact.download(root=download_root_path)
    return Path(download_path)


def build_directory_artifact(
    artifact_name: str,
    artifact_type: ArtifactType,
    dir_path: str | Path,
    *,
    reference: bool = False,
    entry_name: str | None = None,
    max_objects: int | None = None,
) -> wandb.Artifact:
    """Build an artifact containing the contents of a directory.

    Args:
        artifact_name (str): Name of the artifact.
        artifact_type (ArtifactType): Type of artifact.
        dir_path (str | Path): Directory path to include in the artifact.

    Keyword Args:
        reference (bool): Only reference the directory, do not copy contents. Defaults to False.
        entry_name (str | None): Name for directory within the artifact. Defaults to None.
        max_objects (int | None): Max objects to include in the artifact. Defaults to None.

    Returns:
        wandb.Artifact: The generated artifact.
    """
    artifact = wandb.Artifact(name=artifact_name, type=artifact_type)
    if reference:
        # Right now, we are assuming a fixed "file" URI scheme
        # We can pass the URI scheme if necessary later
        artifact.add_reference(
            uri=f"{PathPrefix.FILE.value}{dir_path}",
            max_objects=max_objects,
            name=entry_name,
        )
    else:
        artifact.add_dir(str(dir_path), name=entry_name)
    return artifact


def build_table_artifact(
    artifact_name: str,
    artifact_type: ArtifactType,
    tables: dict[str, pd.DataFrame],
) -> wandb.Artifact:
    """Build an artifact containing one or more table entries.

    Args:
        artifact_name (str): Name of the artifact.
        artifact_type (ArtifactType): Type of artifact.
        tables (dict[str, pd.DataFrame]): Mapping from table name to table data
            in the form of a `pd.DataFrame` object.

    Returns:
        wandb.Artifact: The artifact containing the table(s).
    """
    artifact = wandb.Artifact(artifact_name, type=artifact_type)
    for table_name, table_data in tables.items():
        table = wandb.Table(data=table_data)
        artifact.add(table, name=table_name)
    return artifact



import click

from lm_buddy import LMBuddy
from lm_buddy.cli.utils import parse_config_option
from lm_buddy.configs.jobs import (
    HuggingFaceEvalJobConfig,
    LMHarnessJobConfig,
    PrometheusJobConfig,
    RagasJobConfig,
)


@click.group(name="evaluate", help="Run an LM Buddy evaluation job.")
def group() -> None:
    pass


@group.command("lm-harness", help="Run the lm-harness evaluation job.")
@click.option("--config", type=str)
def lm_harness_command(config: str) -> None:
    config = parse_config_option(LMHarnessJobConfig, config)
    buddy = LMBuddy()
    buddy.evaluate(config)


@group.command("prometheus", help="Run the prometheus evaluation job.")
@click.option("--config", type=str)
def prometheus_command(config: str) -> None:
    config = parse_config_option(PrometheusJobConfig, config)
    buddy = LMBuddy()
    buddy.evaluate(config)


@group.command("ragas", help="Run the ragas evaluation job.")
@click.option("--config", type=str)
def ragas_command(config: str) -> None:
    config = parse_config_option(RagasJobConfig, config)
    buddy = LMBuddy()
    buddy.evaluate(config)


@group.command("huggingface", help="Run the HuggingFace evaluation job.")
@click.option("--config", type=str)
def huggingface_command(config: str) -> None:
    config = parse_config_option(HuggingFaceEvalJobConfig, config)
    buddy = LMBuddy()
    buddy.evaluate(config)



from pathlib import Path
from typing import Any

import wandb
from loguru import logger
from ray import train
from ray.train import CheckpointConfig, RunConfig, ScalingConfig
from ray.train.huggingface.transformers import RayTrainReportCallback, prepare_trainer
from ray.train.torch import TorchTrainer
from transformers import TrainingArguments
from trl import SFTTrainer

from lm_buddy.configs.jobs.finetuning import FinetuningJobConfig
from lm_buddy.jobs.asset_loader import (
    HuggingFaceDatasetLoader,
    HuggingFaceModelLoader,
    HuggingFaceTokenizerLoader,
)
from lm_buddy.jobs.common import FinetuningResult, JobType
from lm_buddy.preprocessing import format_dataset_with_prompt
from lm_buddy.tracking.artifact_utils import (
    ArtifactType,
    build_directory_artifact,
    default_artifact_name,
)
from lm_buddy.tracking.run_utils import WandbResumeMode


def is_tracking_enabled(config: FinetuningJobConfig):
    # Only report to WandB on the rank 0 worker
    # Reference: https://docs.ray.io/en/latest/train/user-guides/experiment-tracking.html
    return config.tracking is not None and train.get_context().get_world_rank() == 0


def load_and_train(config: FinetuningJobConfig):
    # Load the HF assets from configurations
    hf_model_loader = HuggingFaceModelLoader()
    hf_tok_loader = HuggingFaceTokenizerLoader()
    hf_dataset_loader = HuggingFaceDatasetLoader()
    model = hf_model_loader.load_pretrained_model(config.model, config.quantization)
    tokenizer = hf_tok_loader.load_pretrained_tokenizer(config.tokenizer)

    datasets = hf_dataset_loader.load_and_split_dataset(config.dataset)
    if config.dataset.prompt_template is not None:
        for split, dataset in datasets.items():
            datasets[split] = format_dataset_with_prompt(
                dataset=dataset,
                prompt_template=config.dataset.prompt_template,
                output_field=config.dataset.text_field,
            )

    training_args = TrainingArguments(
        output_dir="out",  # Local checkpoint path on a worker
        report_to="wandb" if is_tracking_enabled(config) else "none",
        use_cpu=not config.ray.use_gpu,
        push_to_hub=False,
        disable_tqdm=True,
        logging_dir=None,
        **config.trainer.training_args(),
    )

    peft_config = config.adapter.as_huggingface() if config.adapter else None

    trainer = SFTTrainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        peft_config=peft_config,
        train_dataset=datasets["train"],
        eval_dataset=datasets.get("test"),
        max_seq_length=config.trainer.max_seq_length,
        dataset_text_field=config.dataset.text_field,
    )
    trainer.add_callback(RayTrainReportCallback())
    trainer = prepare_trainer(trainer)
    trainer.train()


def training_function(config_data: dict[str, Any]):
    config = FinetuningJobConfig(**config_data)
    if is_tracking_enabled(config):
        with wandb.init(
            name=config.name,
            resume=WandbResumeMode.NEVER,
            job_type=JobType.FINETUNING,
            **config.tracking.model_dump(),
        ):
            load_and_train(config)
    else:
        load_and_train(config)


def run_finetuning(config: FinetuningJobConfig) -> FinetuningResult:
    # Construct Ray train configurations from input config
    scaling_config = ScalingConfig(
        use_gpu=config.ray.use_gpu,
        num_workers=config.ray.num_workers,
    )
    run_config = RunConfig(
        name=config.name,
        storage_path=config.ray.storage_path,
        checkpoint_config=CheckpointConfig(num_to_keep=1),
    )
    trainer = TorchTrainer(
        train_loop_per_worker=training_function,
        train_loop_config=config.model_dump(),
        scaling_config=scaling_config,
        run_config=run_config,
    )
    result = trainer.fit()
    logger.info(f"Training result: {result}")

    # Create a checkpoint artifact if tracking is enabled and Ray saved a checkpoint
    if result.checkpoint:
        checkpoint_path = Path(result.checkpoint.path) / RayTrainReportCallback.CHECKPOINT_NAME
        checkpoint_artifact = build_directory_artifact(
            artifact_name=default_artifact_name(config.name, ArtifactType.MODEL),
            artifact_type=ArtifactType.MODEL,
            dir_path=checkpoint_path,
            reference=True,
        )
    else:
        checkpoint_path, checkpoint_artifact = None, None

    # Return finetuning result object
    return FinetuningResult(
        artifacts=[checkpoint_artifact] if checkpoint_artifact else [],
        checkpoint_path=checkpoint_path,
        metrics=result.metrics,
        is_adapter=config.adapter is not None,
    )



"""
lm-buddy entrypoint to run evaluations using a Prometheus inference server
see https://github.com/kaistAI/prometheus/blob/main/evaluation/benchmark/run_absolute_scoring.py
"""

import copy
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from datasets import Dataset
from loguru import logger
from openai import OpenAI, OpenAIError
from openai.types import Completion
from tqdm import tqdm

from lm_buddy.configs.huggingface import AutoTokenizerConfig
from lm_buddy.configs.jobs.prometheus import PrometheusJobConfig
from lm_buddy.constants import LM_BUDDY_RESULTS_PATH
from lm_buddy.jobs.asset_loader import (
    HuggingFaceDatasetLoader,
    HuggingFaceModelLoader,
    HuggingFaceTokenizerLoader,
)
from lm_buddy.jobs.common import EvaluationResult
from lm_buddy.jobs.evaluation.conversation import get_conv_template
from lm_buddy.preprocessing import format_dataset_with_prompt
from lm_buddy.tracking.artifact_utils import (
    ArtifactType,
    build_directory_artifact,
    default_artifact_name,
)


@dataclass
class BadResponseError(Exception):
    def __init__(self, message, error=None):
        self.message = message
        self.error = error


def openai_completion(
    config: PrometheusJobConfig, client: OpenAI, engine: str, prompt: str
) -> Completion:
    """Connects to a remote OpenAI-API-compatible Prometheus endpoint
    and returns a Completion holding the model's response.
    """

    return client.completions.create(
        model=engine,
        prompt=prompt,
        best_of=config.prometheus.best_of,
        max_tokens=config.prometheus.max_tokens,
        frequency_penalty=config.prometheus.frequency_penalty,
        temperature=config.prometheus.temperature,
        top_p=config.prometheus.top_p,
    )


def parse_response(config: PrometheusJobConfig, response: Completion) -> tuple[str, str]:
    """Given a Prometheus eval response as returned by the OpenAI API
    endpoint (i.e. in Completion format), extract feedback and score.
    """

    if response is None:
        raise BadResponseError("Server returned an empty response")

    try:
        response_text = response.choices[0].text
        # note: this can raise a ValueError if the message is malformed
        feedback, score = response_text.split("[RESULT]")
        feedback = feedback.strip()
        score = score.strip()
        if score not in [str(s) for s in config.evaluation.scores]:
            raise BadResponseError(f"Score {score} is not in range")
    except (ValueError, BadResponseError) as e:
        raise BadResponseError(f"Server returned a malformed response ({e})", e)

    return feedback, score


def instruction_to_prompt(config: PrometheusJobConfig, instruction: str) -> str:
    """Given some text containing Prometheus instructions, a conversation
    template (e.g. "llama-2") and a system message (e.g. "You are a
    fair evaluator language model"), generate an actual prompt.
    """
    conv = get_conv_template(config.evaluation.conversation_template)
    conv.set_system_message(config.evaluation.conversation_system_message)
    conv.append_message(conv.roles[0], instruction)
    conv.append_message(conv.roles[1], None)
    return conv.get_prompt()


def get_response_with_retries(
    config: PrometheusJobConfig,
    client: OpenAI,
    engine: str,
    prompt: str,
) -> tuple[str, str]:
    current_retry_attempt = 1
    while current_retry_attempt <= config.evaluation.max_retries:
        try:
            response = openai_completion(config, client, engine, prompt)
            feedback, score = parse_response(config, response)
            break
        except (OpenAIError, BadResponseError) as e:
            logger.warning(
                f"{e.message}: "
                f"Retrying ({current_retry_attempt}/{config.evaluation.max_retries})"
            )
            current_retry_attempt += 1
            if current_retry_attempt > config.evaluation.max_retries:
                raise e
    return (feedback, score)


def run_eval(config: PrometheusJobConfig) -> Path:
    # Resolve the engine model
    hf_model_loader = HuggingFaceModelLoader()
    engine_path = hf_model_loader.resolve_asset_path(config.prometheus.inference.engine)

    # Load dataset from W&B artifact
    hf_dataset_loader = HuggingFaceDatasetLoader()
    dataset = hf_dataset_loader.load_dataset(config.dataset)
    if config.dataset.prompt_template is not None:
        dataset = format_dataset_with_prompt(
            dataset, config.dataset.prompt_template, config.dataset.text_field
        )

    # Get the tokenizer
    hf_tok_loader = HuggingFaceTokenizerLoader()
    tokenizer_config = AutoTokenizerConfig(path=config.prometheus.inference.engine)
    tokenizer = hf_tok_loader.load_pretrained_tokenizer(tokenizer_config)

    # Enable / disable tqdm
    dataset_iterable = tqdm(dataset) if config.evaluation.enable_tqdm else dataset

    # Generator that iterates over samples and yields new rows with the prometheus outputs
    def data_generator():
        # Instantiate OpenAI client to speak with the vLLM endpoint
        # Client is non-serializable so must be instantiated internal to this method
        # Reference: https://huggingface.co/docs/datasets/en/troubleshoot#pickling-issues
        client = OpenAI(base_url=config.prometheus.inference.base_url)

        for sample in dataset_iterable:
            # convert instructions from the dataset (`text_field` in a dict) to
            # prompts that prometheus accepts
            prompt = instruction_to_prompt(config, sample[config.dataset.text_field])

            # skip those examples which are too long
            tokenized_prompt = tokenizer(prompt, truncation=False)
            if len(tokenized_prompt["input_ids"]) > 3072:
                logger.warning(f"Skipping row due to prompt exceeding token limit: {prompt=}")
                continue

            # prepare output
            result: dict[str, Any] = copy.deepcopy(sample)
            result["prometheus_output"] = []
            result["prometheus_score"] = []

            for _ in range(config.evaluation.num_answers):
                (feedback, score) = get_response_with_retries(config, client, engine_path, prompt)
                result["prometheus_output"].append(feedback)
                result["prometheus_score"].append(score)

            yield result

    result_dataset = Dataset.from_generator(data_generator)

    # Save dataset to disk
    storage_path = config.evaluation.storage_path or LM_BUDDY_RESULTS_PATH
    result_dataset_path = Path(storage_path) / config.name / "prometheus"
    result_dataset.save_to_disk(str(result_dataset_path))

    return result_dataset_path


def run_prometheus(config: PrometheusJobConfig) -> EvaluationResult:
    # Run eval and store output in local filename
    result_dataset_path = run_eval(config)
    logger.info(f"Prometheus evaluation dataset stored at {result_dataset_path}")

    # Create a directory artifact for the HF dataset
    artifact_name = default_artifact_name(config.name, artifact_type=ArtifactType.DATASET)
    dataset_artifact = build_directory_artifact(
        artifact_name=artifact_name,
        artifact_type=ArtifactType.DATASET,
        dir_path=result_dataset_path,
        reference=False,
    )

    return EvaluationResult(
        artifacts=[dataset_artifact],
        dataset_path=result_dataset_path,
        tables={},
    )



import os
import warnings

from pydantic import model_validator
from wandb.apis.public import Run
from wandb.util import random_string

from lm_buddy.configs.common import LMBuddyConfig


class WandbRunConfig(LMBuddyConfig):
    """Configuration required to log to a W&B run.

    A W&B Run is uniquely identified by the combination of `<entity>/<project>/<id>`.
    The W&B platform will auto-generate values for these variables if they are not provided
    when you initialize a run.

    However, based on how these attributes are passed between jobs it is often necessary
    to know the run ID before initializing a run.
    For this reason, the run ID field is made non-optional and auto-generated locally
    if it is not provided.
    """

    __match_args__ = ("id", "project", "group", "entity")

    id: str
    project: str | None = None
    group: str | None = None
    entity: str | None = None

    @model_validator(mode="before")
    def warn_missing_api_key(cls, values):
        if not os.environ.get("WANDB_API_KEY", None):
            warnings.warn(
                "Cannot find `WANDB_API_KEY` in your environment. "
                "Tracking will fail if a default key does not exist on the Ray cluster."
            )
        return values

    @model_validator(mode="before")
    def ensure_run_id(cls, values):
        if values.get("id", None) is None:
            # Generate an random 8-digit alphanumeric string, analogous to W&B platform
            values["id"] = random_string(length=8)
        return values

    @classmethod
    def from_run(cls, run: Run) -> "WandbRunConfig":
        """Extract environment settings from a W&B Run object.

        Useful when listing runs from the W&B API and extracting their settings for a job.
        """
        # TODO: Can we get the run group from this when it exists?
        return cls(
            project=run.project,
            entity=run.entity,
            id=run.id,
        )

    def wandb_path(self) -> str:
        """String identifier for the asset on the W&B platform."""
        path = "/".join(x for x in [self.entity, self.project, self.id] if x is not None)
        return path

    def env_vars(self) -> dict[str, str]:
        env_vars = {
            "WANDB_RUN_ID": self.id,
            "WANDB_PROJECT": self.project,
            "WANDB_RUN_GROUP": self.group,
            "WANDB_ENTITY": self.entity,
            "WANDB_API_KEY": os.environ.get("WANDB_API_KEY", None),
        }
        return {k: v for k, v in env_vars.items() if v is not None}



import contextlib
import tempfile
from abc import abstractmethod
from pathlib import Path

from pydantic import Field
from pydantic_yaml import parse_yaml_file_as, to_yaml_file

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.wandb import WandbRunConfig
from lm_buddy.paths import AssetPath, PathPrefix


class JobConfig(LMBuddyConfig):
    """Configuration that comprises the entire input to an LM Buddy job.

    This class implements helper methods for de/serializing the configuration from file.

    Currently, there is a 1:1 mapping between job entrypoints and job config implementations,
    but this is not rigidly constrained by the interface. This may change in the future.
    """

    name: str = Field(description="Name of the job.")
    tracking: WandbRunConfig | None = Field(
        default=None,
        description=(
            "Tracking information to associate with the job. "
            "A new run is created with these details."
        ),
    )

    @classmethod
    def from_yaml_file(cls, path: Path | str):
        return parse_yaml_file_as(cls, path)

    def to_yaml_file(self, path: Path | str):
        to_yaml_file(path, self, exclude_none=True)

    @contextlib.contextmanager
    def to_tempfile(self, *, name: str = "config.yaml", dir: str | Path | None = None):
        """Enter a context manager with the config written to a temporary YAML file.

        Keyword Args:
            name (str): Name of the config file in the tmp directory. Defaults to "config.yaml".
            dir (str | Path | None): Root path of the temporary directory. Defaults to None.

        Returns:
            Path to the temporary config file.
        """
        with tempfile.TemporaryDirectory(dir=dir) as tmpdir:
            config_path = Path(tmpdir) / name
            self.to_yaml_file(config_path)
            yield config_path

    @abstractmethod
    def asset_paths(self) -> set[AssetPath]:
        """Return a set of all `AssetPath` fields on this config."""
        pass

    def artifact_paths(self) -> set[AssetPath]:
        """Return a set of all W&B artifact paths on this config."""
        return {x for x in self.asset_paths() if x.startswith(PathPrefix.WANDB)}



import warnings

import torch
from accelerate import Accelerator
from datasets import Dataset, DatasetDict, load_dataset, load_from_disk
from loguru import logger
from peft import PeftConfig
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)
from transformers.models.auto.modeling_auto import (
    MODEL_FOR_CAUSAL_LM_MAPPING_NAMES,
    MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING_NAMES,
)

from lm_buddy.configs.huggingface import (
    AutoModelConfig,
    AutoTokenizerConfig,
    DatasetConfig,
    QuantizationConfig,
)
from lm_buddy.paths import AssetPath, PathPrefix, strip_path_prefix
from lm_buddy.tracking.artifact_utils import get_artifact_directory, get_artifact_from_api


class HuggingFaceAssetLoader:
    """Helper class for loading HuggingFace assets from LM Buddy configurations.

    The base class provides implementation for resolving the asset path.
    Resolution depends both on the source and on the artifact type.
    TODO: implement resolution for non-dataset artifacts

    """

    def resolve_asset_path(self, path: AssetPath) -> str:
        """Resolve an `AssetPath` to a loadable string path.

        W&B paths are resolved to file paths given the artifact manifest.
        The returned string has its `PathPrefix` stripped away..
        """
        raw_path = strip_path_prefix(path)
        if path.startswith(
            (PathPrefix.FILE, PathPrefix.HUGGINGFACE, PathPrefix.OPENAI, PathPrefix.MISTRAL)
        ):
            return raw_path
        elif path.startswith(PathPrefix.WANDB):
            artifact = get_artifact_from_api(raw_path)
            return str(get_artifact_directory(artifact))
        elif path.startswith(PathPrefix.S3):
            if isinstance(self, HuggingFaceDatasetLoader):
                return path
            else:
                raise ValueError(f"S3 support for {self.__class__} is not available yet")
        else:
            raise ValueError(f"Unable to resolve asset path from {path}.")


class HuggingFaceModelLoader(HuggingFaceAssetLoader):
    """Helper class for loading HuggingFace models from LM Buddy configurations."""

    def resolve_peft_and_pretrained(self, path: AssetPath) -> tuple[str, str | None]:
        """Helper method for determining if a path corresponds to a PEFT model.

        A PEFT model contains an `adapter_config.json` in its directory.
        If this file can be loaded, we know the path is a for a PEFT model.
        If not, we assume the provided path corresponds to a base HF model.

        Args:
            path (AssetPath): Path for the asset with its `PathPrefix` present

        Returns:
            Tuple of (base model path, optional PEFT path)
        """
        # We don't know if the checkpoint is adapter weights or merged model weights
        # Try to load as an adapter and fall back to the checkpoint containing the full model
        resolved_path = self.resolve_asset_path(path)
        try:
            peft_config = PeftConfig.from_pretrained(resolved_path)
            return peft_config.base_model_name_or_path, resolved_path
        except ValueError as e:
            warnings.warn(
                f"Unable to load model as adapter: {e}. "
                "This is expected if the checkpoint does not contain adapter weights."
            )
            return resolved_path, None

    def load_pretrained_config(
        self,
        config: AutoModelConfig,
    ) -> PretrainedConfig:
        """Load a `PretrainedConfig` from the model configuration.

        An exception is raised if the HuggingFace repo does not contain a `config.json` file.
        """
        config_path = self.resolve_asset_path(config.path)
        return AutoConfig.from_pretrained(
            pretrained_model_name_or_path=config_path, trust_remote_code=config.trust_remote_code
        )

    def load_pretrained_model(
        self,
        config: AutoModelConfig,
        quantization: QuantizationConfig | None = None,
    ) -> PreTrainedModel:
        """Load a `PreTrainedModel` with optional quantization from the model configuration.

        An exception is raised if the HuggingFace repo does not contain a `config.json` file.

        TODO(RD2024-87): This fails if the checkpoint only contains a PEFT adapter config
        """
        device_map, bnb_config = None, None
        if quantization is not None:
            bnb_config = quantization.as_huggingface()
            # When quantization is enabled, model must all be on same GPU to work with DDP
            # If a device_map is not specified we will get accelerate errors downstream
            # Reference: https://github.com/huggingface/accelerate/issues/1840#issuecomment-1683105994
            current_device = (
                Accelerator().local_process_index if torch.cuda.is_available() else "cpu"
            )
            device_map = {"": current_device}
            logger.info(f"Setting model device_map = {device_map} to enable quantization")

        # TODO: HuggingFace has many AutoModel classes with different "language model heads"
        #   Can we abstract this to load with any type of AutoModel class?
        model_path = self.resolve_asset_path(config.path)

        # load config first to get the model type
        model_config = self.load_pretrained_config(config)

        if getattr(model_config, "model_type") in MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING_NAMES:
            automodel_class = AutoModelForSeq2SeqLM
        elif getattr(model_config, "model_type") in MODEL_FOR_CAUSAL_LM_MAPPING_NAMES:
            automodel_class = AutoModelForCausalLM
        else:
            logger.info("Model type not supported. Trying AutoModelForCausalLM")
            automodel_class = AutoModelForCausalLM

        return automodel_class.from_pretrained(
            pretrained_model_name_or_path=model_path,
            trust_remote_code=config.trust_remote_code,
            torch_dtype=config.torch_dtype,
            quantization_config=bnb_config,
            device_map=device_map,
        )


class HuggingFaceTokenizerLoader(HuggingFaceAssetLoader):
    """Helper class for loading HuggingFace tokenizers from LM Buddy configurations."""

    def load_pretrained_tokenizer(self, config: AutoTokenizerConfig) -> PreTrainedTokenizer:
        """Load a `PreTrainedTokenizer` from the model configuration.

        An exception is raised if the HuggingFace repo does not contain a `tokenizer.json` file.
        """
        tokenizer_path = self.resolve_asset_path(config.path)
        tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=tokenizer_path,
            trust_remote_code=config.trust_remote_code,
            use_fast=config.use_fast,
            model_max_length=config.mod_max_length,
        )
        if tokenizer.pad_token_id is None:
            # Pad token required for generating consistent batch sizes
            tokenizer.pad_token_id = tokenizer.eos_token_id
        return tokenizer


class HuggingFaceDatasetLoader(HuggingFaceAssetLoader):
    """Helper class for loading HuggingFace datasets from LM Buddy configurations."""

    def load_dataset(self, config: DatasetConfig) -> Dataset:
        """Load a HuggingFace `Dataset` from the dataset configuration.

        This method always returns a single `Dataset` object.
        When loading from HuggingFace directly, the `Dataset` is for the provided split.
        When loading from disk, the saved files must be for a dataset else an exception is raised.
        """
        dataset_path = self.resolve_asset_path(config.path)
        # Dataset loading requires a different method if from a HF vs. disk
        if config.path.startswith(PathPrefix.HUGGINGFACE):
            return load_dataset(dataset_path, split=config.split)
        else:
            match load_from_disk(dataset_path):
                case Dataset() as dataset:
                    return dataset
                case other:
                    raise ValueError(
                        "LM Buddy currently only supports loading `Dataset` objects from disk, "
                        f"instead found a {type(other)}."
                    )

    def load_and_split_dataset(self, config: DatasetConfig) -> DatasetDict:
        """Load a HuggingFace dataset and optionally perform a train/test split.

        The split is performed when a `test_size` is specified on the configuration.
        """
        match self.load_dataset(config):
            case Dataset() as dataset if config.test_size is not None:
                # We need to specify a fixed seed to load the datasets on each worker
                # Under the hood, HuggingFace uses `accelerate` to create a data loader shards
                # If the datasets are not seeded here, the ordering will be inconsistent
                # TODO: Get rid of this when data is loaded once outside of Ray workers
                split_seed = config.seed or 0
                return dataset.train_test_split(test_size=config.test_size, seed=split_seed)
            case dataset:
                return DatasetDict({"train": dataset})



from typing import Any

import lm_eval
import pandas as pd
import torch
from lm_eval.models.huggingface import HFLM
from lm_eval.models.openai_completions import OpenaiCompletionsLM
from loguru import logger

from lm_buddy.configs.huggingface import AutoModelConfig
from lm_buddy.configs.jobs.lm_harness import LMHarnessJobConfig, LocalChatCompletionsConfig
from lm_buddy.jobs.asset_loader import HuggingFaceModelLoader
from lm_buddy.jobs.common import EvaluationResult
from lm_buddy.tracking.artifact_utils import (
    ArtifactType,
    build_table_artifact,
    default_artifact_name,
)


def get_per_task_dataframes(
    results: dict[str, dict[str, Any]],
) -> dict[str, pd.DataFrame]:
    """Create a `pd.DataFrame` of numeric metrics for each evaluation task.

    This is necessary because artifact tables must have a single datatype for each column.

    lm-harness returns mostly numeric values, but there are also some misc string-valued metrics.
    Filtering down to only numeric values allows us to produce a valid table artifact.
    """
    task_dataframes = {}
    for task_name, data in results.items():
        numeric_rows = [(k, v) for k, v in data.items() if isinstance(v, int | float)]
        task_dataframes[task_name] = pd.DataFrame(data=numeric_rows, columns=["metric", "value"])
    return task_dataframes


def load_harness_model(config: LMHarnessJobConfig) -> HFLM | OpenaiCompletionsLM:
    # Instantiate the lm-harness LM class based on the provided model config type
    hf_model_loader = HuggingFaceModelLoader()
    match config.model:
        case AutoModelConfig() as model_config:
            model_path, peft_path = hf_model_loader.resolve_peft_and_pretrained(model_config.path)
            quantization_kwargs: dict[str, Any] = (
                config.quantization.model_dump() if config.quantization else {}
            )
            # TODO: Fix this up by passing in the instantiated model directly
            return HFLM(
                pretrained=model_path,
                tokenizer=model_path,
                peft=peft_path,
                device="cuda" if torch.cuda.device_count() > 0 else "cpu",
                trust_remote_code=model_config.trust_remote_code,
                dtype=config.model.torch_dtype if config.model.torch_dtype else "auto",
                **quantization_kwargs,
            )

        case LocalChatCompletionsConfig() as local_config:
            engine_path = hf_model_loader.resolve_asset_path(local_config.inference.engine)
            return OpenaiCompletionsLM(
                model=engine_path,
                base_url=local_config.inference.base_url,
                tokenizer_backend=local_config.tokenizer_backend,
                truncate=local_config.truncate,
                max_gen_toks=local_config.max_tokens,
            )

        case _:
            raise ValueError(f"Unexpected model config type: {type(config.model)}")


def run_lm_harness(config: LMHarnessJobConfig) -> EvaluationResult:
    logger.info(
        f"Running lm-harness evaluation with configuration:\n {config.model_dump_json(indent=2)}"
    )

    llm = load_harness_model(config)
    eval_results = lm_eval.simple_evaluate(
        model=llm,
        tasks=config.evaluation.tasks,
        batch_size=config.evaluation.batch_size,
        num_fewshot=config.evaluation.num_fewshot,
        limit=config.evaluation.limit,
        log_samples=False,
    )
    logger.info(f"Obtained evaluation results: {eval_results}")

    # Create an artifact containing eval tables
    result_tables = get_per_task_dataframes(eval_results["results"])
    artifact_name = default_artifact_name(config.name, ArtifactType.EVALUATION)
    table_artifact = build_table_artifact(
        artifact_name=artifact_name,
        artifact_type=ArtifactType.EVALUATION,
        tables=result_tables,
    )

    return EvaluationResult(
        tables=result_tables,
        artifacts=[table_artifact],
        dataset_path=None,
    )



import os
from abc import abstractmethod

import torch
from loguru import logger
from mistralai.client import MistralClient
from openai import OpenAI, OpenAIError
from openai.types import Completion
from transformers import pipeline

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.jobs.hf_evaluate import HuggingFaceEvalJobConfig
from lm_buddy.configs.vllm import VLLMCompletionsConfig
from lm_buddy.jobs.asset_loader import HuggingFaceModelLoader, HuggingFaceTokenizerLoader


class BaseModelClient:
    """
    Abstract class for a model client, used to provide a uniform interface
    (currentnly just a simple predict method) to models served in different
    ways (e.g. HF models loaded locally, OpenAI endpoints, vLLM inference
    servers, llamafile).
    """

    @abstractmethod
    def __init__(self, model: str, config: LMBuddyConfig):
        """
        Used to initialize the model / inference service.
        """
        pass

    @abstractmethod
    def predict(self, prompt: str) -> str:
        """
        Given a prompt, return a prediction.
        """
        pass


class SummarizationPipelineModelClient(BaseModelClient):
    """
    Model client for the huggingface summarization pipeline
    (model is loaded locally).
    """

    def __init__(self, model: str, config: HuggingFaceEvalJobConfig):
        self._config = config

        hf_tokenizer_loader = HuggingFaceTokenizerLoader()
        self._tokenizer = hf_tokenizer_loader.load_pretrained_tokenizer(config.tokenizer)

        self._summarizer = pipeline(
            "summarization",
            model=model,
            tokenizer=self._tokenizer,
            device=0 if torch.cuda.is_available() else -1,
            truncation=True,
        )

    def predict(self, prompt):
        # summarizer output is a list (1 element in this case) of dict with key = "summary_text"
        # TODO: bring summarizer parameters out at some point (not needed at the moment)
        pred = self._summarizer(prompt, min_length=30, do_sample=False)
        return pred[0]["summary_text"]


class HuggingFaceModelClient(BaseModelClient):
    """
    Model client for HF models (model is loaded locally, both Seq2SeqLM
    and CausalLM are supported).
    - Provide model path to load the model locally
    - Make sure you add quantization details if the model is too large
    - Optionally, add a tokenizer (the one matching the specified model name is the default)
    """

    def __init__(self, model: str, config: HuggingFaceEvalJobConfig):
        self._config = config
        self._device = "cuda" if torch.cuda.is_available() else "cpu"

        hf_model_loader = HuggingFaceModelLoader()
        hf_tokenizer_loader = HuggingFaceTokenizerLoader()
        self._model = hf_model_loader.load_pretrained_model(config.model).to(self._device)
        self._tokenizer = hf_tokenizer_loader.load_pretrained_tokenizer(config.tokenizer)

    def predict(self, prompt):
        inputs = self._tokenizer(prompt, truncation=True, padding=True, return_tensors="pt").to(
            self._device
        )
        generated_ids = self._model.generate(**inputs, max_new_tokens=256)
        return self._tokenizer.batch_decode(generated_ids, skip_special_tokens=True)[0]


class APIModelClient(BaseModelClient):
    """General model client for APIs."""

    def __init__(self, config: VLLMCompletionsConfig):
        self._config = config

        hf_model_loader = HuggingFaceModelLoader()
        self._engine = hf_model_loader.resolve_asset_path(config.inference.engine)
        self._system = config.inference.system_prompt

    @abstractmethod
    def _chat_completion(
        self,
        config: VLLMCompletionsConfig,
        client: OpenAI | MistralClient,
        prompt: str,
        system: str,
    ) -> Completion:
        """Connects to the API and returns a chat completion holding the model's response."""
        pass

    def _get_response_with_retries(
        self,
        config: VLLMCompletionsConfig,
        prompt: str,
    ) -> tuple[str, str]:
        current_retry_attempt = 1
        max_retries = 1 if config.inference.max_retries is None else config.inference.max_retries
        while current_retry_attempt <= max_retries:
            try:
                response = self._chat_completion(self._config, self._client, prompt, self._system)
                break
            except OpenAIError as e:
                logger.warning(f"{e.message}: Retrying ({current_retry_attempt}/{max_retries})")
                current_retry_attempt += 1
                if current_retry_attempt > max_retries:
                    raise e
        return response

    def predict(self, prompt):
        response = self._get_response_with_retries(self._config, prompt)

        return response.choices[0].message.content


class OpenAIModelClient(APIModelClient):
    """
    Model client for models served via openai-compatible API.
    For OpenAI models:
    - The base_url is fixed
    - Choose an engine name (see https://platform.openai.com/docs/models)
    - Customize the system prompt if needed

    For compatible models:
    - Works with local/remote vLLM-served models and llamafiles
    - Provide base_url and engine
    - Customize the system prompt if needed
    """

    def __init__(self, model: str, config: VLLMCompletionsConfig):
        super().__init__(config)
        self._client = OpenAI(base_url=model)

    def _chat_completion(
        self,
        config: VLLMCompletionsConfig,
        client: OpenAI,
        prompt: str,
        system: str = "You are a helpful assisant.",
    ) -> Completion:
        """Connects to a remote OpenAI-API-compatible endpoint
        and returns a chat completion holding the model's response.
        """

        return client.chat.completions.create(
            model=self._engine,
            messages=[{"role": "system", "content": system}, {"role": "user", "content": prompt}],
            max_tokens=config.max_tokens,
            frequency_penalty=config.frequency_penalty,
            temperature=config.temperature,
            top_p=config.top_p,
        )


class MistralModelClient(APIModelClient):
    """
    Model client for models served via Mistral API.
    - The base_url is fixed
    - Choose an engine name (see https://docs.mistral.ai/getting-started/models/)
    - Customize the system prompt if needed
    """

    def __init__(self, model: str, config: VLLMCompletionsConfig):
        super().__init__(config)
        self._client = MistralClient(api_key=os.environ["MISTRAL_API_KEY"])

    def _chat_completion(
        self,
        config: VLLMCompletionsConfig,
        client: MistralClient,
        prompt: str,
        system: str = "You are a helpful assisant.",
    ) -> Completion:
        """Connects to a Mistral endpoint
        and returns a chat completion holding the model's response.
        """

        return client.chat(
            model=self._engine,
            messages=[{"role": "system", "content": system}, {"role": "user", "content": prompt}],
            max_tokens=config.max_tokens,
            temperature=config.temperature,
            top_p=config.top_p,
        )



import functools
import time

from loguru import logger


def timer(func):
    """
    Decorator which times the execution of the wrapped func.
    Execution time is logged and also returned together with func's returned value
    (output will be a tuple).
    """

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        tic = time.perf_counter()
        value = func(*args, **kwargs)
        toc = time.perf_counter()
        elapsed_time = toc - tic
        logger.info(f"Elapsed time for {func.__name__}: {elapsed_time:0.4f} seconds")
        return value, elapsed_time

    return wrapper_timer



import dataclasses
from typing import Any

from peft import PeftConfig, PeftType, TaskType
from pydantic import field_validator, model_validator
from transformers import BitsAndBytesConfig

from lm_buddy.configs.common import LMBuddyConfig, SerializableTorchDtype
from lm_buddy.paths import AssetPath, PathPrefix

DEFAULT_TEXT_FIELD: str = "text"


class AutoModelConfig(LMBuddyConfig):
    """Settings passed to a HuggingFace AutoModel instantiation.

    The model to load can either be a HuggingFace repo or an artifact reference on W&B.
    """

    path: AssetPath
    trust_remote_code: bool = False
    torch_dtype: SerializableTorchDtype | None = None


class AutoTokenizerConfig(LMBuddyConfig):
    """Settings passed to a HuggingFace AutoTokenizer instantiation."""

    path: AssetPath
    trust_remote_code: bool | None = None
    use_fast: bool | None = None
    mod_max_length: int | None = None


class DatasetConfig(LMBuddyConfig):
    """Settings passed to load a HuggingFace text dataset.

    The dataset can either contain a single text column named by the `text_field` parameter,
    or a `prompt_template` can be provided to format columns of the dataset as the `text_field`.
    """

    path: AssetPath
    text_field: str = DEFAULT_TEXT_FIELD
    prompt_template: str | None = None
    split: str | None = None
    test_size: float | None = None
    seed: int | None = None

    @model_validator(mode="after")
    def validate_split_if_huggingface_path(cls, config: "DatasetConfig"):
        """
        Ensure a  `split` is provided when loading a HuggingFace dataset directly from HF Hub.
        This makes it such that the `load_dataset` function returns the type `Dataset`
        instead of `DatasetDict`, which makes some of the downstream logic easier.
        """
        if config.split is None and config.path.startswith(PathPrefix.HUGGINGFACE):
            raise ValueError(
                "A `split` must be specified when loading a dataset directly from HuggingFace."
            )
        return config


class AdapterConfig(LMBuddyConfig, extra="allow"):
    """Configuration containing PEFT adapter settings.

    The type of adapter is controlled by the required `peft_type` field,
    which must be one of the allowed values from the PEFT `PeftType` enumeration.
    Extra arguments are allowed and are passed down to the HuggingFace `PeftConfig`
    class determined by the `peft_type` argument.

    The `task_type` for the adapter is also required.
    By default, this is set to `TaskType.CAUSAL_LM`
    which is appropriate for causal language model finetuning.
    See the allowed values in the PEFT `TaskType` enumeration.
    """

    peft_type: PeftType
    task_type: TaskType = TaskType.CAUSAL_LM

    @staticmethod
    def _get_peft_config_class(peft_type: PeftType) -> type[PeftConfig]:
        # Internal import to avoid bringing the global variable from peft into module scope
        from peft.mapping import PEFT_TYPE_TO_CONFIG_MAPPING

        return PEFT_TYPE_TO_CONFIG_MAPPING[peft_type]

    @field_validator("peft_type", "task_type", mode="before")
    def sanitize_enum_args(cls, x):
        if isinstance(x, str):
            x = x.strip().upper()
        return x

    @model_validator(mode="after")
    def validate_adapter_args(cls, config: "AdapterConfig"):
        peft_type = config.peft_type

        # PeftConfigs are standard dataclasses so can extract their allowed field names
        adapter_cls = cls._get_peft_config_class(peft_type)
        allowed_fields = {x.name for x in dataclasses.fields(adapter_cls)}

        # Filter fields to those found on the PeftConfig
        extra_fields = config.model_fields_set.difference(allowed_fields)
        if extra_fields:
            raise ValueError(f"Unknown arguments for {peft_type} adapter: {extra_fields}")

        return config

    def as_huggingface(self) -> PeftConfig:
        adapter_cls = self._get_peft_config_class(self.peft_type)
        return adapter_cls(**self.model_dump())


class QuantizationConfig(LMBuddyConfig):
    """Basic quantization settings to pass to training and evaluation jobs.

    Note that in order to use BitsAndBytes quantization on Ray,
    you must ensure that the runtime environment is installed with GPU support.
    This can be configured by setting the `entrypoint_num_gpus > 0` when submitting a job
    to the cluster.
    """

    load_in_8bit: bool | None = None
    load_in_4bit: bool | None = None
    bnb_4bit_quant_type: str = "fp4"
    bnb_4bit_compute_dtype: SerializableTorchDtype | None = None

    def as_huggingface(self) -> BitsAndBytesConfig:
        return BitsAndBytesConfig(
            load_in_4bit=self.load_in_4bit,
            load_in_8bit=self.load_in_8bit,
            bnb_4bit_compute_dtype=self.bnb_4bit_compute_dtype,
            bnb_4bit_quant_type=self.bnb_4bit_quant_type,
        )


class TrainerConfig(LMBuddyConfig):
    """Configuration for a HuggingFace trainer/training arguments.

    This mainly encompasses arguments passed to the HuggingFace `TrainingArguments` class,
    but also contains some additional parameters for the `Trainer` or `SFTTrainer` classes.
    """

    max_seq_length: int | None = None
    num_train_epochs: float | None = None
    per_device_train_batch_size: int | None = None
    per_device_eval_batch_size: int | None = None
    learning_rate: float | None = None
    weight_decay: float | None = None
    gradient_accumulation_steps: int | None = None
    gradient_checkpointing: bool | None = None
    evaluation_strategy: str | None = None
    eval_steps: float | None = None
    logging_strategy: str | None = None
    logging_steps: float | None = None
    save_strategy: str | None = None
    save_steps: int | None = None

    def training_args(self) -> dict[str, Any]:
        """Return the arguments to the HuggingFace `TrainingArguments` class."""
        return self.model_dump(exclude={"max_seq_length"}, exclude_none=True)



from lm_buddy.configs.jobs.common import JobConfig
from lm_buddy.configs.jobs.finetuning import FinetuningJobConfig
from lm_buddy.configs.jobs.hf_evaluate import HuggingFaceEvalJobConfig
from lm_buddy.configs.jobs.lm_harness import LMHarnessJobConfig
from lm_buddy.configs.jobs.prometheus import PrometheusJobConfig
from lm_buddy.configs.jobs.ragas import RagasJobConfig

EvaluationJobConfig = (
    LMHarnessJobConfig | PrometheusJobConfig | RagasJobConfig | HuggingFaceEvalJobConfig
)

__all__ = [
    "JobConfig",
    "FinetuningJobConfig",
    "LMHarnessJobConfig",
    "PrometheusJobConfig",
    "RagasJobConfig",
    "HuggingFaceEvalJobConfig",
    "EvaluationJobConfig",
]



from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.paths import AssetPath


class InferenceServerConfig(LMBuddyConfig):
    """Generic configuration for an externally hosted inference endpoint.

    The inference server is defined by an endpoint with the provided `base_url`.
    The model `engine` powering the endpoint can also be specified
    to provide further metadata about the model used for inference
    (e.g., for generating W&B artifact lineages).

    Note: This configuration is intended to be generic and not bound to the interface
    of any specific training/evaluation framework. See `LocalChatCompletionConfig`
    or `vLLMCompletionsConfig` for intended usage alongside a third-party framework.
    """

    base_url: str
    engine: AssetPath

    # optional system prompt to be used by default in chat completions
    system_prompt: str | None = None

    # max number of retries when communication with server fails
    max_retries: int | None = None


class VLLMCompletionsConfig(LMBuddyConfig):
    """Configuration for a vLLM-based completions service

    The "local-chat-completions" model is powered by a self-hosted inference server,
    specified as an `InferenceServerConfig`. Additional arguments are also provided
    to control the tokenizer type and generation parameters.

    Note that this is just a subset of the parameters allowed by a vLLM server (see
    https://github.com/vllm-project/vllm/blob/main/vllm/sampling_params.py). If we
    choose to use this configuration to cover for more use cases, it will make sense
    to add the other supported configuration parameters too.
    """

    inference: InferenceServerConfig
    # vLLM-specific params
    best_of: int | None = None
    max_tokens: int | None = None
    frequency_penalty: float | None = None
    temperature: float | None = None
    top_p: float | None = None
    top_k: int | None = None
